import numpy as np
import scipy as sp
from scipy.ndimage import filters
from skimage import measure
from skimage.util import view_as_windows


def _is_intermediate(img, footprint):
    min_ = filters.minimum_filter(img, footprint=footprint)
    max_ = filters.maximum_filter(img, footprint=footprint)
    is_intermediate_value = (min_ <= img) * (img <= max_)
    is_intermediate_value = is_intermediate_value.astype(int)
    return is_intermediate_value[2:-2, 2:-2]


def is_intermediate(img, isotropic=True):
    if isotropic:
        footprint = np.array([[0, 1, 0], [1, 0, 1],
                              [0, 1, 0]])[:, :, None].astype(bool)
        is_intermediate_value = _is_intermediate(img, footprint)
    else:
        footprint_h = np.array([[1, 0, 1]])[:, :, None].astype(bool)
        intermediate_h = _is_intermediate(img, footprint_h)
        intermediate_v = _is_intermediate(img, footprint_h.transpose(1, 0, 2))
        is_intermediate_value = .5 * (intermediate_h + intermediate_v)
    return is_intermediate_value


def create_sliding_windows(x, W, stride):
    """
    :param x: array-like, shape (Y, X, C)
    :return windows: np.ndarray, shape (n_windows, W, W, C)
    :return cuts_y: np.ndarray, shape (n_windows,): Y coordinate of the centre of each window
    :return cuts_x: np.ndarray, shape (n_windows,)
    """
    Y, X, C = x.shape
    windows = view_as_windows(x, (W, W, 3), stride).reshape(-1, W, W, 3)
    # windows = np.lib.stride_tricks.sliding_window_view(
    # x, (W, W), axis=(0, 1))[::stride, ::stride]
    cuts_y = np.arange(W // 2, Y - W // 2 + 1)[::stride] - 0.5
    cuts_x = np.arange(W // 2, X - W // 2 + 1)[::stride] - 0.5
    # windows = windows.reshape(-1, C, W, W).transpose(0, 2, 3, 1)
    return windows, cuts_y, cuts_x


def find_grid(x):
    """
    :param x: array-like, shape (n, W, W, C)
    """
    n, W_Y, W_X, C = x.shape
    count_diag_xggx = np.sum(x[:, ::2, 1::2, 1], axis=(1, 2)) + np.sum(
        x[:, 1::2, ::2, 1], axis=(1, 2))
    count_diag_gxxg = np.sum(x[:, ::2, ::2, 1], axis=(1, 2)) + np.sum(
        x[:, 1::2, 1::2, 1], axis=(1, 2))
    diff_diag = 2 * (count_diag_xggx - count_diag_gxxg) / (W_Y * W_X)
    best_diag = (diff_diag > 0).astype(int)  # 0 if XGGX, 1 if GXXG

    count_rggb = np.sum(x[:, ::2, ::2, 0] + x[:, 1::2, 1::2, 2], axis=(1, 2))
    count_grbg = np.sum(x[:, ::2, 1::2, 0] + x[:, 1::2, ::2, 2], axis=(1, 2))
    count_gbrg = np.sum(x[:, 1::2, ::2, 0] + x[:, ::2, 1::2, 2], axis=(1, 2))
    count_bggr = np.sum(x[:, 1::2, 1::2, 0] + x[:, ::2, ::2, 2], axis=(1, 2))

    diff_rggb_bggr = 2 / (W_Y * W_X) * (count_rggb - count_bggr)
    diff_grbg_gbrg = 2 / (W_Y * W_X) * (count_grbg - count_gbrg)

    diff_grid = best_diag * diff_grbg_gbrg + (1 - best_diag) * diff_rggb_bggr
    best_rb = (diff_grid > 0).astype(int)  # 0: RGGB or GRBG, 1:BGGR or GBRG
    best_grid = 2 * best_diag + best_rb  # 0: RGGB, 1: BGGR, 2: GRBG, 3: GBRG

    return best_grid, best_diag, diff_grid, diff_diag, diff_rggb_bggr, diff_grbg_gbrg


def find_forgeries(img, W=32, stride=4, isotropic=False, threshold=.1):
    Y, X, C = img.shape
    img = img[:Y - Y % 2, :X - X % 2]
    intermediate = is_intermediate(img, isotropic=isotropic)
    windows, cuts_y, cuts_x = create_sliding_windows(intermediate, W, stride)

    def to_fullsize(data, method, fill=None):
        interpolator = sp.interpolate.RegularGridInterpolator(
            (cuts_y, cuts_x),
            data.reshape(len(cuts_y), len(cuts_x)),
            method=method,
            bounds_error=False,
            fill_value=fill)
        target_coords = np.asarray(
            np.meshgrid(np.arange(0, Y), np.arange(0, X)))
        result = interpolator(target_coords.reshape(2, -1).T).reshape(X, Y).T
        return result

    cuts_y += 2
    cuts_x += 2
    cY, cX = len(cuts_y), len(cuts_x)
    grid, diag, diff, diff_diag, diff_rggb_bggr, diff_grbg_gbrg = find_grid(
        windows)

    # diff_normalized = np.abs(diff) / np.max(np.abs(diff))
    # diff_diag_normalized = np.abs(diff_diag) / np.max(np.abs(diff_diag))

    main_grid, main_diag = find_grid(intermediate[None])[:2]
    main_grid = main_grid[0]
    main_diag = main_diag[0]

    diag[diff_diag == 0] = -1
    grid[diff == 0] = -1

    abs_diff = np.abs(diff)
    abs_diff_diag = np.abs(diff_diag)

    # connected maximum
    labels_diag, n_regions = measure.label(diag.reshape(cY, cX),
                                           background=main_diag,
                                           return_num=True)
    labels_diag = labels_diag.ravel()
    inconsistent_diag_con = np.zeros_like(diff_diag)
    for i_region in range(1, n_regions + 1):
        mask = labels_diag == i_region
        max_ = np.max(abs_diff_diag[mask])
        inconsistent_diag_con[mask] = max_

    labels_grid, n_regions = measure.label(grid.reshape(cY, cX),
                                           background=main_grid,
                                           return_num=True)
    labels_grid = labels_grid.ravel()
    inconsistent_grid_con = np.zeros_like(diff)
    for i_region in range(1, n_regions + 1):
        mask = labels_grid == i_region
        max_ = np.max(abs_diff[mask])
        inconsistent_grid_con[mask] = max_

    # vals = abs_diff_diag * (diag != main_diag)
    # vals = vals.reshape(cY, cX)
    # inconsistent_diag_thr = apply_hysteresis_threshold(vals, 1e-20,
    # threshold).ravel()
    # inconsistent_grid_thr = np.zeros_like(inconsistent_diag_thr)
    # for g in range(4):
    # if g != main_grid:
    # vals = (abs_diff * (grid == g)).reshape(cY, cX)
    # inconsistent_grid_thr += apply_hysteresis_threshold(
    # vals, 1e-20, threshold).ravel()
    # inconsistent_diag_thr = inconsistent_diag_thr.astype(float)
    # inconsistent_grid_thr = inconsistent_grid_thr.astype(float)

    # coords = np.stack(np.meshgrid(cuts_x, cuts_y)).reshape(2, -1).T
    # inconsistent_diag_seg = np.zeros(coords.shape[0])
    # inconsistent_grid_seg = np.zeros(coords.shape[0])

    # dbscan = DBSCAN(eps=dbscan_eps * W,
    # min_samples=dbscan_min_weight * (W // stride)**2)

    # c = coords[diag != main_diag]
    # w = abs_diff_diag[diag != main_diag]
    # labels = dbscan.fit_predict(c, sample_weight=w) >= 0
    # inconsistent_diag_seg[diag != main_diag] = labels

    # for g in range(4):
    # if g != main_grid:
    # c = coords[grid == g]
    # w = np.abs(diff)[grid == g]
    # labels = dbscan.fit_predict(c, sample_weight=w) >= 0
    # inconsistent_grid_seg[grid == g] = labels

    inconsistent_diag_raw = (diag != main_diag) * (diff_diag != 0)
    # inconsistent_diag_soft = inconsistent_diag * diff_diag_normalized

    inconsistent_grid_raw = (grid != main_grid) * (diff != 0) * (diff_diag !=
                                                                 0)
    inconsistent_full_raw = np.maximum(inconsistent_diag_raw,
                                       inconsistent_grid_raw)
    # inconsistent_grid_soft = inconsistent_grid * diff_normalized

    inconsistent_full_con = np.maximum(inconsistent_diag_con,
                                       inconsistent_grid_con)

    inconsistent_grid = inconsistent_grid_con > threshold
    inconsistent_diag = inconsistent_diag_con > threshold
    inconsistent_full = inconsistent_full_con > threshold

    grid = to_fullsize(grid, 'nearest')
    diag = to_fullsize(diag, 'nearest')
    inconsistent_grid = to_fullsize(inconsistent_grid, 'nearest', 0)
    inconsistent_diag = to_fullsize(inconsistent_diag, 'nearest', 0)
    inconsistent_full = to_fullsize(inconsistent_full, 'nearest', 0)
    inconsistent_grid_con = to_fullsize(inconsistent_grid_con, 'linear', 0)
    inconsistent_diag_con = to_fullsize(inconsistent_diag_con, 'linear', 0)
    inconsistent_full_con = to_fullsize(inconsistent_full_con, 'linear', 0)
    inconsistent_grid_raw = to_fullsize(inconsistent_grid_raw, 'nearest', 0)
    inconsistent_diag_raw = to_fullsize(inconsistent_diag_raw, 'nearest', 0)
    inconsistent_full_raw = to_fullsize(inconsistent_full_raw, 'nearest', 0)
    diff = to_fullsize(diff, 'linear', 0)
    diff_rggb_bggr = to_fullsize(diff_rggb_bggr, 'linear', 0)
    diff_grbg_gbrg = to_fullsize(diff_grbg_gbrg, 'linear', 0)
    diff_diag = to_fullsize(diff_diag, 'linear', 0)

    return grid, diag, main_grid, inconsistent_grid, inconsistent_diag, inconsistent_full, inconsistent_grid_con, inconsistent_diag_con, inconsistent_full_con, inconsistent_grid_raw, inconsistent_diag_raw, inconsistent_full_raw, diff, diff_diag, diff_rggb_bggr, diff_grbg_gbrg
